#include <getopt.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <stdio.h>
#include <pthread.h>
#include <stdlib.h>
#include <errno.h>
#include <signal.h>
#include <sys/socket.h>
#include <netdb.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <linux/kd.h>
#include <linux/keyboard.h>
#include <linux/vt.h>
#include "uinput.h"
#include "ringbuffer.h"
#include "event.h"
#include "load_config.h"

#define EVENT_TIMEOUT 200       /* event timeout in ms */
#define CONN_TIMEOUT 2          /* connection timeeout in seconds */
#define MAX_COUNT 256           /* maximum events during one read */
#define KEY_OFFSET 500          /* offset to indicate special keys */
#define CONSOLE_DEV "/dev/console"
#define MAX_STR 255

#define USAGE  " \
kbdsniffd version 0.4 \n \
keyboard sniffer daemon - to by-pass keyboard events \n \
Author: Marco Skambraks <marco@suse.de> \n\n \
-n --nodaemon - don't fork\n \
-i --info - display device information \n \
-s --scan - scan for first keyboard device number \n \
-h --help - display this help \n \
-b --basedir - device basedir (/dev/input) \n \
-c --config - config file (/etc/kbdsniffd.conf) \n \
-e --event - event device (0) \n \
-u --uinput - uinput device (uinput) \n \
-p --port - port for incoming connections (8899)\n \
-k --key - authentication key (default) \n \
-v --verbose - be verbose \n\n "

struct option long_options[] = {
  {"nodaemon", 0, 0, 'n'},
  {"verbose", 0, 0, 'v'},
  {"basedir", 0, 0, 'b'},
  {"event", 1, 0, 'e'},
  {"uinput", 1, 0, 'u'},
  {"info", 0, 0, 'i'},
  {"help", 0, 0, 'h'},
  {"port", 1, 0, 'p'},
  {"key", 1, 0, 'k'},
  {"config", 1, 0, 'c'},
  {"scan", 0, 0, 's'},
  {0, 0, 0, 0}
};

void *grabber_thread ();
int keycode_trans (int keycode);
void close_all ();

/* handler for signal "isgpipe" */
void sigpipe (int signum);

/* termination handler */
void terminate (int signum);

/* socket communication */
void handle_connection ();

/* helper function to send message over the socket */
int sock_send (int sock_id, char *str);

/* helper function - parse the config file */
int load_config (char *file);

/* global variables */
int verbose = 0;
char authkey[MAX_STR] = "default";
int port = 8899;
char basedir[MAX_STR] = "/dev/input";
char uinput_dev[MAX_STR] = "uinput";
int event_dev_num = -1;

int sock_listen = 0;
int con_fd = 0;
volatile sig_atomic_t grabber_loop = 1;
volatile sig_atomic_t main_loop = 1;
volatile sig_atomic_t conn_loop = 1;
volatile int sniffer_key = 0;

pthread_mutex_t key_buf_mutex = PTHREAD_MUTEX_INITIALIZER;
pthread_mutex_t loop_mutex = PTHREAD_MUTEX_INITIALIZER;

int main (int argc, char **argv)
{
  pthread_t thread_id;
  int c = 0;
  int do_fork = 1;
  struct hostent *hp = NULL, *gethostbyname ();

  char conf[MAX_STR] = "/etc/kbdsniffd.conf";
  int info = 0;
  struct sockaddr_in sock_addr;
  int sock_new = 0;
  int sock_reuse = 1;
  int sock_ret = 0;

  /* try to load config from default path */
  load_config (conf);

  /* parse options */
  while (1)
   {
     int option_index = 0;

     c =
       getopt_long (argc, argv, "nsb:c:hive:k:u:p:", long_options,
                    &option_index);
     if (c == -1)
       break;

     switch (c)
      {
      case 0:
        printf (" option %s", long_options[option_index].name);
        if (optarg)
          printf (" with arg %s", optarg);
        printf ("\n");
        break;

      case 'h':
        printf ("%s", USAGE);
        return 0;

      case 'i':
        info = 1;
        break;

      case 'b':
        strcpy (basedir, optarg);
        break;
      case 'p':
        port = atoi (optarg);
        break;
      case 'k':
        strcpy (authkey, optarg);
        break;
      case 'u':
        strcpy (uinput_dev, optarg);
        break;
      case 'c':
        strcpy (conf, optarg);
        load_config (conf);
        break;
      case 'e':
        event_dev_num = atoi (optarg);
        break;

      case 'v':
        verbose = 1;
        break;

      case 's':
       {
         int tmp = 0;

         do
          {
            tmp = event_scan (basedir, tmp, 23);
            if (tmp >= 0)
             {
               printf ("eventt=%d \n", tmp);
               tmp++;
             }

          }
         while (tmp >= 0);
       }
        return 0;

      case 'n':
        do_fork = 0;
        break;

      case '?':
        return 2;
        break;

      }
   }

  if (optind < argc)
   {
     fprintf (stderr, " non - option: ");
     return 2;
     while (optind < argc)
       fprintf (stderr, "%s", argv[optind++]);
     fprintf (stderr, "\n");

     return 2;
   }

  if (do_fork)
   {
     switch (fork ())
      {
      case -1:
        fprintf (stderr, "error: fork failed \n");
        return 1;
      case 0:
        setsid ();
        /* become a daemon */
        break;
      default:
        return 0;
      }

   }

  if (event_dev_num == -1)
    event_dev_num = event_scan (basedir, 0, 23);
  else if (event_dev_num == -2)
   {
     int tmp = 0;

     /* look for the last keyboard dev */
     do
      {
        tmp = event_scan (basedir, tmp, 23);
        if (tmp >= 0)
         {
           event_dev_num = tmp;
           tmp++;
         }

      }
     while (tmp >= 0);
   }

  if (verbose)
    printf ("basedir=%s uinput=%s evnum=%d - localhost:%d\n", basedir,
            uinput_dev, event_dev_num, port);

  if (uinput_init (basedir, uinput_dev))
   {
     fprintf (stderr, "%s", USAGE);
     fprintf (stderr,
              "can't open %s/%s \nplease make sure that the module uinput is loaded\n",
              basedir, uinput_dev);
     return 1;
   }

  if (event_init (basedir, event_dev_num))
   {
     fprintf (stderr, "%s", USAGE);
     fprintf (stderr, "can't open %s/event%d\n", basedir, event_dev_num);
     close_all ();
     return 2;
   }

  if (verbose || info)
   {
     event_info ();
     if (info)
      {
        close_all ();
        return 0;
      }

   }

  if ((con_fd = open (CONSOLE_DEV, O_RDWR)) < 0)
   {
     close_all ();
     fprintf (stderr, "can't open %s\n", CONSOLE_DEV);
     return 3;
   }

  /* set socket properties */
  sock_listen = socket (AF_INET, SOCK_STREAM, IPPROTO_TCP);
  sock_ret =
    setsockopt (sock_listen, SOL_SOCKET, SO_REUSEADDR, &sock_reuse,
                sizeof (sock_reuse));
  if (sock_ret < 0)
   {
     close_all ();
     fprintf (stderr, "can't set socket option\n");
     return 4;
   }

  sock_addr.sin_family = AF_INET;
  sock_addr.sin_port = htons (port);

  if ((hp = gethostbyname ("localhost")) == NULL)
   {
     close_all ();
     fprintf (stderr, "can't get host (localhost)\n");
     return 5;
   }
  else
    memcpy (&sock_addr.sin_addr, hp->h_addr, hp->h_length);

  sock_ret =
    bind (sock_listen, (struct sockaddr *) &sock_addr, sizeof (sock_addr));

  if (sock_ret < 0)
   {
     close_all ();
     fprintf (stderr, "can't bind socket\n");
     return 7;
   }

  if (listen (sock_listen, 0) < 0)
   {
     close_all ();
     fprintf (stderr, "liten failed\n");
     return 8;
   }

  signal (SIGPIPE, sigpipe);
  signal (SIGINT, terminate);
  signal (SIGTERM, terminate);
  signal (SIGCHLD, SIG_IGN);
  signal (SIGHUP, SIG_IGN);

  while (main_loop)
   {

     if (verbose)
       printf ("listening on port=%d\n", port);

     sock_new = accept (sock_listen, NULL, NULL);
     if (sock_new >= 0)
      {
        grabber_loop = 1;
        conn_loop = 1;
        key_buf_init ();
        pthread_create (&thread_id, NULL, grabber_thread, NULL);
        pthread_detach (thread_id);
        sched_yield ();
        handle_connection (sock_new);

        pthread_mutex_lock (&loop_mutex);
        grabber_loop = 0;
        pthread_mutex_unlock (&loop_mutex);
        close (sock_new);
        while (sniffer_key)
          usleep (1000);
      }

   }

  close (sock_listen);
  close_all ();

  if (verbose)
    printf ("%s finished\n", argv[0]);

  return 0;
}

int keycode_trans (int keycode)
{
  struct kbentry key;
  int type = 0;

  switch (keycode)
   {
   case 1:
   case 14:
   case 15:
   case 28:
   case 57:
     return (keycode + KEY_OFFSET);
   }

  key.kb_table = 0;
  key.kb_index = keycode;
  key.kb_value = 0;
  if (ioctl (con_fd, KDGKBENT, &key) < 0)
    return 0;

  type = KTYP (key.kb_value) & 0x0f;

  if (type == KT_LETTER || type == KT_LATIN)
    return KVAL (key.kb_value);
  else
    return (keycode + KEY_OFFSET);
}

void *grabber_thread ()
{
  int i = 0, rd = 0;
  int sniff_on = 0;
  struct input_event ev[MAX_COUNT];
  int con_fd;                   /* console fd */
  int con_nr = 0;               /* current console */
  int con_mode = 0;             /* console mode 0 = text - other values
                                   indicate none-text-mode */
  struct vt_stat vterm;
  int event_err = 0;

  if (verbose)
    printf ("grabber thread started pid=%d\n", (unsigned int) getpid ());

  /* wait until a sniffer key is set or grabber is disabled */
  while (grabber_loop && !sniffer_key)
    usleep (1000);

  if (verbose)
    printf ("switch to grabbing mode - key=%d\n", sniffer_key);

  /* do a dummy read to drop buffered keys */
  while (event_get (ev, MAX_COUNT, 100) != -2)
    usleep (1000);

  event_grab (1);

  con_fd = open ("/dev/console", O_RDWR);
  while (grabber_loop)
   {

     rd = event_get (ev, MAX_COUNT, EVENT_TIMEOUT);
     if (rd == -3)
       event_err = 1;

     for (i = 0; i < rd; i++)
      {
        if (ev[i].type == 1 && ev[i].code == sniffer_key && ev[i].value > 0
            && !con_mode)
          sniff_on = 1;

        /* if not in sniffer mode send the keys back to the system otherwise
           write the keys to a buffer */

        if (!sniff_on)
          uinput_key (ev[i]);
        else if (ev[i].type == 1 && ev[i].code != sniffer_key && ev[i].value)
         {
           pthread_mutex_lock (&key_buf_mutex);
           key_put_buf (keycode_trans (ev[i].code));
           pthread_mutex_unlock (&key_buf_mutex);
         }

        /* disable sniffer mode if the sniffer key is released */
        if (ev[i].type == 1 && ev[i].code == sniffer_key && !ev[i].value)
          sniff_on = 0;
      }

     if (!sniff_on)
      {
        /* check if console has changed */
        ioctl (con_fd, VT_GETSTATE, &vterm);
        if (vterm.v_active != con_nr)
         {
           int mode_fd;
           char tmpstr[MAX_STR];

           sprintf (tmpstr, "/dev/tty%d", vterm.v_active);
           mode_fd = open (tmpstr, O_RDWR);
           ioctl (mode_fd, KDGETMODE, &con_mode);
           close (mode_fd);
           con_nr = vterm.v_active;
         }                      /* if - changed */

      }

     /* check for event-errors - if true = fallback to the first dev */
     if (event_err)
      {
        event_close ();
        event_dev_num = event_scan (basedir, 0, 23);
        event_init (basedir, event_dev_num);
        memset (ev, 0, sizeof (ev));

        /* do a dummy read to drop buffered keys */
        while (event_get (ev, MAX_COUNT, 100) != -2)
          usleep (1000);

        event_grab (1);

        event_err = 0;

      }                         /* if error */
   }

  if (con_fd >= 0)
    close (con_fd);
  event_grab (0);
  sniffer_key = 0;
  if (verbose)
    printf ("grabber thread finished pid=%d\n", (unsigned int) getpid ());

  return 0;
}

void close_all ()
{

  grabber_loop = 0;
  uinput_close ();
  event_close ();
  if (con_fd >= 0)
    close (con_fd);
}

void handle_connection (int sock_id)
{
  struct timeval timeout;
  int auth = 0;
  fd_set select_set;
  char tmpstr[MAX_STR] = "";

  if (verbose)
    printf ("entering connection loop\n");

  while (conn_loop)
   {
     int ret = 0;

     FD_ZERO (&select_set);
     FD_SET (sock_id, &select_set);
     timeout.tv_sec = CONN_TIMEOUT;
     timeout.tv_usec = 0;

     memset (tmpstr, 0, sizeof (tmpstr));

     ret = select (sock_id + 1, &select_set, NULL, NULL, &timeout);
     if (ret > 0)
      {
        ret = recv (sock_id, tmpstr, MAX_STR, MSG_DONTWAIT);

        if (ret <= 0)
         {
           pthread_mutex_lock (&loop_mutex);
           conn_loop = 0;
           grabber_loop = 0;
           auth = 0;
           pthread_mutex_unlock (&loop_mutex);
           continue;
         }
        /* check string */
        if (!strncmp (tmpstr, "auth ", 5))
         {
           tmpstr[strlen (tmpstr) - 2] = 0;
           if (!strcmp (tmpstr + 5, authkey))
            {
              auth = 1;
              conn_loop = sock_send (sock_id, "OK\r\n");
            }
           else
            {
              sock_send (sock_id, "bye\r\n");
              conn_loop = 0;
              auth = 0;
            }

         }
        else if (!strcmp (tmpstr, "get\r\n") && auth)
         {
           char key_str[10] = "";
           int key = 0;

           if (key_getbufcnt ())
            {
              pthread_mutex_lock (&key_buf_mutex);
              key_get_buf (&key);
              pthread_mutex_unlock (&key_buf_mutex);
              sprintf (key_str, "%03d\r\n", key);
            }
           else
             strcpy (key_str, "000\r\n");

           conn_loop = sock_send (sock_id, key_str);
         }
        else if (!strncmp (tmpstr, "setkey ", 7) && auth)
         {
           int key = 0;

           key = atoi (tmpstr + 7);
           if (key > KEY_OFFSET)
            {
              pthread_mutex_lock (&loop_mutex);
              sniffer_key = key - KEY_OFFSET;
              conn_loop = sock_send (sock_id, "OK\r\n");
              pthread_mutex_unlock (&loop_mutex);

            }
           else
             conn_loop =
               sock_send (sock_id, "error: this is not a valid key\r\n");

         }

        else if (!strcmp (tmpstr, "quit\r\n"))
         {
           pthread_mutex_lock (&loop_mutex);
           conn_loop = 0;
           grabber_loop = 0;
           auth = 0;
           pthread_mutex_unlock (&loop_mutex);
         }
        else if (!auth)
          conn_loop = sock_send (sock_id, "error: permission denied\r\n");

      }
     else
      {
        sock_send (sock_id, "bye\r\n");
        pthread_mutex_lock (&loop_mutex);
        conn_loop = 0;
        grabber_loop = 0;
        auth = 0;
        pthread_mutex_unlock (&loop_mutex);
      }

   }

  if (verbose)
    printf ("connection finished\n");

}

int sock_send (int sock_id, char *str)
{
  int ret = 0;
  fd_set select_set;

  FD_ZERO (&select_set);
  FD_SET (sock_id, &select_set);
  struct timeval timeout;

  timeout.tv_sec = 1;
  timeout.tv_usec = 0;
  if (select (sock_id + 1, NULL, &select_set, NULL, &timeout) <= 0)
    return 0;

/* send the string including the \0 character */
  ret = send (sock_id, str, strlen (str), MSG_DONTWAIT);

  if (ret != (int) (strlen (str)))
   {
     if (verbose)
       fprintf (stderr, "sock_send: %d\n", ret);

     return 0;
   }

  return 1;
}
void sigpipe (int signum)
{
  pthread_mutex_lock (&loop_mutex);
  conn_loop = 0;
  grabber_loop = 0;
  pthread_mutex_unlock (&loop_mutex);
  signal (signum, sigpipe);
}

void terminate (int signum)
{
  pthread_mutex_lock (&loop_mutex);
  conn_loop = 0;
  grabber_loop = 0;
  main_loop = 0;
  pthread_mutex_unlock (&loop_mutex);
  fcntl (sock_listen, F_SETFL, O_NONBLOCK);
  signal (signum, terminate);
}
int load_config (char *file)
{
  char tmpstr[MAX_STR] = "";

  if (confread (file))
    return 1;

  confsearch (strcpy (tmpstr, "key"));
  if (tmpstr[0])
    strcpy (authkey, tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "port"));
  if (atoi (tmpstr) > 0)
    port = atoi (tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "basedir"));
  if (tmpstr[0])
    strcpy (basedir, tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "uinput"));
  if (tmpstr[0])
    strcpy (uinput_dev, tmpstr);

  memset (tmpstr, 0, MAX_STR);

  confsearch (strcpy (tmpstr, "event_dev_num"));
  if (atoi (tmpstr) >= -2)
    event_dev_num = atoi (tmpstr);

  return 0;
}
